%{
#include    "lex.h"

Token_Type Token;
int line_number = 1;
%}

whitespace          [ \t]

letter              [a-zA-Z]
digit               [0-9]
underscore          "_"
letter_or_digit     ({letter}|{digit})
underscored_tail    ({underscore}{letter_or_digit}+)
identifier          ({letter}{letter_or_digit}*{underscored_tail}*)

operator            [-+*/]
separator           [;,(){}]

%%

{digit}+                {return INTEGER;}
{identifier}            {return IDENTIFIER;}
{operator}|{separator}  {return yytext[0];}
#[^#\n]*#?              {/* ignore comment */}
{whitespace}            {/* ignore whitespace */}
\n                      {line_number++;}
.                       {return ERRONEOUS;}

%%

void start_lex(void) {}

void get_next_token(void) {
    Token.class = yylex();
    if (Token.class == 0) {
        Token.class = EoF; Token.repr = "<EoF>"; return;
    }
    Token.pos.line_number = line_number;
    Token.repr = strdup(yytext);
}

int yywrap(void) {return 1;}
